import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from pandas_ext.common.utils import fs


def to_parquet(df: pd.DataFrame, path: str, compression:str='snappy', **kwargs) -> None:
    """Convert dataframe to parquet file in path."""
    return pq.write_table(
        pa.Table.from_pandas(df),
        fs(path, 'wb'),
        compression=compression,
        flavor='spark',
        **kwargs
    )


def read_parquet(path:str, **kwargs) -> pd.DataFrame:
    """Read parquet table locally or from s3."""
    return pq.read_table(fs(path, 'rb'), **kwargs).to_pandas()



"""gdrive handlers for easy read/write to Gdrive."""
from collections import namedtuple
from io import BytesIO
from json import JSONDecodeError
from os import environ as env, path, remove
from tempfile import gettempdir

import pandas as pd
import requests

PANDAS_MIMETYPES = {
    'text/csv': pd.read_csv,
    'application/vnd.ms-excel': pd.read_excel,
    'application/vnd.openxmlformats-officedocument.'
    'spreadsheetml.sheet': pd.read_excel,
    'application/json': pd.read_json,
}

PANDAS_EXTENSIONS = dict(
    csv='to_csv',
    xls='to_excel',
    xlsx='to_excel',
    json='to_json',
)


def _get_endpoint_payload():
    """Create route and key headers for gdrive API."""
    endpoint = env['GDRIVE_URL']
    key = env['GDRIVE_KEY']
    headers = {'x-api-key': key} if key else {}
    return dict(headers=headers, route=endpoint)


def read_gdrive(url: str, raw=False, **kwargs) -> object:
    """Given a url to a csv in gdrive, return namedtuple.

    Args:
        url: gdrive link url
        **kwargs: Passed to read_csv, read_excel, or read_json

    Returns:
        object: Returns dataframe if file mimetype in PANDAS_MIMETYPES
            else file-like object for further processing

    """
    payload = _get_endpoint_payload()
    route = payload['route'] + '/read'
    params = dict(url=url)
    response = requests.get(
        route,
        headers=payload['headers'],
        params=params,
        stream=True,
    )

    mimetype = response.headers['content-type'].split(';')[0]

    if mimetype in PANDAS_MIMETYPES and not raw:
        with BytesIO(response.content) as fh:
            return PANDAS_MIMETYPES[mimetype](fh, **kwargs)
    else:
        return BytesIO(response.content)


def to_gdrive(
    file_name: str,
    folder_id: str,
    data=None,
    to_gapp=False,
    **kwargs
) -> str:
    """Send data to gdrive with a given file_name.ext.

    To send any file that doesn't come from a dataframe, make sure to save it
    in /tmp folder. to_gdrive  will know to look there for file_name param.

    Args:
        file_name: Filename.ext stored in /tmp folder.
        folder_id: folder_id associated with meta.folder_id
        data: (:obj:`pd.DataFrame`, optional): Pandas Dataframe, if applicable
            Binary files like PDF or TXT can be sent through to_gdrive with
            data set to None.
        **kwargs: kwargs to pass to to_csv, to_excel, or to_json

    Returns:
        Google Drive alternate web link for preview.

    """
    payload = _get_endpoint_payload()
    route = payload['route'] + '/write'

    ext = file_name.split('.')[-1].lower()
    tmp_file = path.join(gettempdir(), file_name)

    if (
        isinstance(data, pd.DataFrame) and
        ext in PANDAS_EXTENSIONS and
        not path.exists(tmp_file)
    ):
        getattr(data, PANDAS_EXTENSIONS[ext])(tmp_file, **kwargs)

    if not path.exists(tmp_file):
        raise OSError('%s not found' % tmp_file)

    ifile = dict(
        file=(file_name, open(tmp_file, 'rb'))
    )

    data = dict(
        folder_id=folder_id,
        to_gapp=to_gapp
    )
    response = requests.post(
        route,
        headers=payload['headers'],
        files=ifile,
        data=data,
    )
    remove(tmp_file)
    try:
        return response.json()
    except JSONDecodeError:
        return response.content


def gdrive_metadata(url: str, fetch_all=False) -> object:
    """Given a gdrive url (file or folder), return all metadata.

    Args:
        url: gdrive link url
        fetch_all: Whether or not to fetch all metadata or not.
            Defaults to False.

    Returns:
        namedtuple if fetch_all is False else dictionary of all metadata

    """
    payload = _get_endpoint_payload()
    route = payload['route'] + '/metadata'
    params = dict(url=url)

    response = requests.get(
        route,
        headers=payload['headers'],
        params=params
    )

    metadata = response.json()
    # metadata = {meta[0]: meta[1] for meta in response.json()}

    meta_fields = [
        'mimeType',
        'fileExtension',
        'lastModifyingUser',
        'title',
        'parents',
        'fileSize',
        'alternateLink',
    ]

    try:
        metadata['folder_id'] = (
            None if 'parents' not in metadata
            else metadata['parents'][0]['id']
        )
    except IndexError:
        raise('The file must reside in a folder that is shared with '
              'pydrive@namely.com.')

    if 'lastModifyingUser' in metadata:
        metadata['last_mod_by_email'] = (
            metadata['lastModifyingUser']['emailAddress']
        )
        del metadata['lastModifyingUser']

    if not fetch_all:
        metadata = {
            k: v
            for k, v in metadata.items()
            if k in meta_fields +
            ['folder_id', 'last_mod_by_email']
        }
        del metadata['parents']
        Metadata = namedtuple('MetaData', metadata.keys())
        return Metadata(**metadata)

    return metadata



"""SFDC handler for reading from SFDC."""
from concurrent.futures import ThreadPoolExecutor
from os import environ as env

from jinja2 import Environment
import pandas as pd
import requests


def _get_endpoint_payload():
    """Create route and key headers for API."""
    endpoint = env['SFDC_URL']
    key = env['SFDC_KEY']
    headers = {'x-api-key': key}
    return dict(headers=headers, route=endpoint)


def read_sfdc(sql: str, sql_params=None) -> str:
    """Given sql and params, run the sql."""
    payload = _get_endpoint_payload()
    route = payload['route'] + '/sfdc/query/v20.0'

    sql_params = sql_params if sql_params else {}
    rendered_sql = Environment().from_string(sql).render(**sql_params)

    params = dict(sql=rendered_sql)
    response = requests.get(
        route,
        headers=payload['headers'],
        params=params,
    )
    response.raise_for_status()
    return pd.read_json(response.json())


def sfdc_metadata(sobject: str, fetch_all=False) -> str:
    """Given a sfdc object, return its metadata."""
    payload = _get_endpoint_payload()
    route = payload['route'] + f'/sfdc/metadata/{sobject}/v20.0'

    response = requests.get(
        route,
        headers=payload['headers'],
    )

    response.raise_for_status()
    metadata = response.json()

    if fetch_all:
        return metadata

    return [meta['name'] for meta in metadata['fields']]


def patch_sfdc(sf_url, data):
    """Sync patch sfdc."""
    payload = _get_endpoint_payload()
    route = payload['route'] + '/sfdc/patch/v20.0'

    params = dict(sf_url=sf_url, data=data)
    response = requests.patch(
        route,
        headers=payload['headers'],
        params=params,
    )
    response.raise_for_status()
    return response.json()


def async_patch_sfdc(data):
    """Asynchronously patch sfdc."""
    payload = _get_endpoint_payload()
    route = payload['route'] + '/sfdc/patch/v20.0'

    params = [
        dict(sf_url=items['sf_url'], data=items['patch_data'])
        for items in data
    ]
    threaded_params = (dict(
        url=route,
        headers=payload['headers'],
        params=param

    ) for param in params)
    print(list(threaded_params))

    with ThreadPoolExecutor(max_workers=25) as executor:
        pass
        responses = executor.map(
            requests.patch, **(lambda x: x for x in threaded_params))
    print(list(responses))
    return responses



import pandas as pd
import s3fs

from pandas_ext.common.utils import is_s3_path


def to_csv(df: pd.DataFrame, path: str, **kwargs) -> None:
    """Given a df, write it to s3 if necessary."""
    if is_s3_path(path):
        bytes_to_write = df.to_csv(None, **kwargs).encode()

        s3 = s3fs.S3FileSystem()
        with s3.open(path, 'wb') as dest:
            return dest.write(bytes_to_write)

    return df.to_csv(path, **kwargs)




"""Write a pandas dataframe to xml."""
import pandas as pd
from xml.sax.saxutils import escape


def to_xml(df, filename=None, mode='w'):
    """Write df to xml."""
    def row_to_xml(row):
        """Row by row."""
        xml = ['<item>']
        for i, col_name in enumerate(row.index):
            raw = row.iloc[i]
            if not isinstance(raw, float) and '&' in raw:
                print(raw, type(raw))
            value = escape(raw) if isinstance(raw, str) else raw

            xml.append('  <field name="{0}">{1}</field>'
                       .format(col_name, value)
                       )
        xml.append('</item>')
        return '\n'.join(xml)

    res = '<?xml version="1.0" encoding="UTF-8"?>\n'
    res += '\n'.join(df.apply(row_to_xml, axis=1))

    if filename is None:
        return res
    with open(filename, mode) as f:
        f.write(res)


pd.DataFrame.to_xml = to_xml



import datetime

from functools import partial
from enum import auto, Enum
from os import path

import s3fs


class Mode(Enum):
    rb = auto()
    wb = auto()


def is_s3_path(path: str):
    return path.startswith('s3://')


def fs(path:str, mode:Mode='rb'):
    """Opens file locally or via s3 depending on path string."""
    s3 = s3fs.S3FileSystem()
    return partial(
        {True: s3.open, False: open}[is_s3_path(path)],
        mode=mode
    )(path)


def today():
    return str(datetime.date.today())

def now():
    return str(datetime.datetime.now())



"""Versioning kept here."""
__version__ = '0.4.1'
__license__ = "MIT"

__title__ = "pandas_ext"
__description__ = "Python Pandas extensions for pandas dataframes"

__author__ = "Rich Fernandez, Sean Massot, Brian Tenazas, Luke Orland"
__email__ = "devs@newsela.com"

__uri__ = "https://github.com/newsela/pandas_ext"

import pandas as _pd

from .amazon_spectrum import to_spectrum
from .csv import to_csv
from .gdrive import read_gdrive, to_gdrive
from .parquet import read_parquet, to_parquet
from .sfdc import (
    read_sfdc, sfdc_metadata, patch_sfdc, async_patch_sfdc,
)
from .sql import read_sql, list_backends


read_csv = _pd.read_csv

# Don't pollute the namespace with the module names:
del amazon_spectrum
del common
del csv
del gdrive
del parquet
del sfdc
del sql
del _pd
